@@ -15,6 +15,10 @@ use parent qw(Debian::Debhelper::Buildsystem);
 
 use constant MAX_BUILD_LOOP => 3;
 
+sub NPMIGNOREDEFAULT {
+    $ENV{NPMIGNOREDEFAULT} || '/usr/share/pkg-js-tools/npmignore.default';
+}
+
 # echo, touch and ls are here to help tests
 my @known_build_commands = (qw(grunt echo touch ls));
 
@@ -45,13 +49,13 @@ sub DESCRIPTION {
 
 # Regexp compiled from dh_nodejs/dh_root_files.excluded
 my $rootFilesIgnored =
-qr/^(?:c(?:o(?:ntribute|pying)|hange(?:log|s?)|akefile)|a(?:ut(?:otest\.watchr|hors?)|ppveyor\.yml)|g(?:ulpfile(?:\.babel)?|runt(?:file)?)\.js|b(?:inding\.gyp|ench.*\.js|ower\.json)|.*\.(?:m(?:arkdown|d)|pdf|txt)|r(?:ollup\.config\.js|eadme)|t(?:sconfig\.json|ests?\.js)|l(?:icen[cs]e.*|erna\.json)|(?:Docker|make)file|package-lock\.json|yarn\.lock|history|\..*)$/i;
+qr/^(?:(?:g(?:ulpfile\.(?:babel\.j|(?:t|j))|runt(?:file)?\.j)|(?:rollup[\.\-].*config|karma\.conf|.mocharc)\.j)s|b(?:abel\.config\.js(?:on)?|inding\.gyp|ench.*\.js|ower\.json)|c(?:o(?:mponent\.json|ntribute|pying)|hange(?:log|s?)|akefile)|a(?:ut(?:otest\.watchr|hors?)|va\.config\.js|ppveyor\.yml)|j(?:est\.config\.js|sl\.node\.conf)|.*\.(?:m(?:arkdown|d)|pdf|txt)|t(?:sconfig\.json|ests?\.js)|l(?:icen[cs]e.*|erna\.json)|\.(?:babelrc(?:.js)?|.*)|(?:Docker|make)file|package-lock\.json|yarn\.lock|history)$/i;
 my $filesIgnored =
-qr/^(?:\.(?:git(?:ignore)?|npmignore|eslint.*)|.*\.(?:c(?:offee|\+\+|cp?)?|def|mk|h)|(?:[jm]|binding.M)akefile)$/i;
+qr/^(?:\.(?:(?:wafpickle-|eslint|_).*|npm(?:ignore|rc)|lock-wscript|gitignore)|.*\.(?:c(?:offee|\+\+|cp?)?|orig|def|mk|h)|(?:[jm]|binding.M)akefile|readme(?:\.(?:txt|md))?|package-lock\.json|npm-debug\.log|yarn\.lock)$/i;
 
 # Regexp compiled from dh_nodejs/dh_dirs.excluded
 my $dirIgnored =
-qr{^(?:(?:node_module|(.*/)?.dep)s|(?:example|test|doc)s?|bench|\..*)(?:/.*)?}i;
+qr{^(?:(.*/)?(?:\.(?:_.*|git|svn|hg)|.(?:DS_Store|deps)|__[^/]*__|fixtures?|CVS)|(?:archived-packag|node_modul)es|(?:example|test|doc)s?|bench|\..*)(?:/.*)?}i;
 
 # Extensions that overrides $rootFilesIgnored if set in package.json#files
 my $authorizedExt = qr/\.(?:js)$/;
@@ -80,8 +84,8 @@ sub main_package {
     return ( $main_p ? $main_p : () ) if defined $main_p;
     return '.' unless -e 'debian/nodejs/main';
     my @lines = $self->open_file('debian/nodejs/main');
-    error 'Malformed debian/nodejs/main'     if @lines > 1;
-    error "Missing main directory $lines[0]" if @lines and !-d $lines[0];
+    error 'Malformed debian/nodejs/main'       if @lines > 1;
+    warning "Missing main directory $lines[0]" if @lines and !-d $lines[0];
     $main_p = $lines[0] || '';
     return @lines;
 }
@@ -282,6 +286,7 @@ sub resolve_command {
 
 sub dolink {
     my ( $self, $src, $dest ) = @_;
+    return if $src eq $dest;
     my $dir = $dest;
     $self->doit_in_builddir( 'mkdir', '-p', $dir )
       if $dir =~ s#/[^/]*$##
@@ -341,6 +346,7 @@ sub install_module {
         } $self->open_file("debian/nodejs/$dir/install");
         return;
     }
+    my $npmignore = -e "$dir/.npmignore" ? "$dir/.npmignore" : NPMIGNOREDEFAULT;
     if ( -e "debian/nodejs/$dir/files" ) {
         @files =
           $self->readFilesField( $dir,
@@ -361,10 +367,20 @@ sub install_module {
               . " in $dir/package.json\nPlease report this bug\n";
         }
         $noFilesField = 0;
+        push @files,
+          grep { $_ ne "$dir/package.json" } $self->readFilesField(
+            $dir,
+            ( map { s/^/!/; s/^\!\!//; $_ } ( $self->open_file($npmignore) ) )
+          );
     }
     else {
         print qq{No "files" field in $dir/package.json, install all files\n};
         @files = ($dir);
+        push @files,
+          grep { $_ ne "$dir/package.json" } $self->readFilesField(
+            $dir,
+            ( map { s/^/!/; s/^\!\!//; $_ } ( $self->open_file($npmignore) ) )
+          );
     }
     my @dest;
     my $mainFile =
@@ -378,6 +394,10 @@ sub install_module {
 
     foreach my $p (@files) {
         my $pattern;
+        my $not = 0;
+        if ( $p =~ s/\!// ) {
+            $not = 1;
+        }
         if ( $p =~ m/[\*|\{]/ ) {
             ( $p, $pattern ) = $self->pattern($p);
         }
@@ -440,16 +460,22 @@ sub install_module {
                 );
 
                 # Install files unless one exclusion test match
-                unless ( grep { $_ } @tests ) {
-                    push @dest, [ $d, $File::Find::name ];
-                    $foundMain = 1 if $File::Find::name eq $mainFile;
+                if ($not) {
+                    @dest = grep { $_->[1] ne $File::Find::name } @dest;
                 }
+                else {
+                    unless ( grep { $_ } @tests ) {
+                        push @dest, [ $d, $File::Find::name ];
+                        $foundMain = 1 if $File::Find::name eq $mainFile;
+                    }
 
-                # Debug
-                elsif ( DEBUG and not DEBUGFILE ) {
-                    warning "# FILE $File::Find::name\n";
-                    for ( my $i = 0 ; $i < @tests ; $i++ ) {
-                        warning "#     Test$i: " . ( $tests[$i] || 0 ) . "\n";
+                    # Debug
+                    elsif ( DEBUG and not DEBUGFILE ) {
+                        warning "# FILE $File::Find::name\n";
+                        for ( my $i = 0 ; $i < @tests ; $i++ ) {
+                            warning "#     Test$i: "
+                              . ( $tests[$i] || 0 ) . "\n";
+                        }
                     }
                 }
                 if ( DEBUGFILE and $_ eq DEBUGFILE ) {
@@ -477,7 +503,8 @@ EOF
         if ( -e $mainFile ) {
             warning
               "package.json#main file not installed, forcing it ($mainFile)\n";
-            push @dest, [ "", $mainFile ];
+            my $ldir = ( $mainFile =~ m#^(?:$dir/?)?(.*)/# ? $1 : "" );
+            push @dest, [ $ldir, $mainFile ];
         }
     }
     foreach (@dest) {
@@ -504,14 +531,21 @@ sub filesFieldToList {
 sub readFilesField {
     my ( $self, $dir, @lines ) = @_;
     my @files = map {
+        my $prefix = '';
         $_ = $self->toJs( $dir, $_ );
+        my $nowarn;
+        if (s/^!//) {
+            $prefix = '!';
+            $nowarn = 1;
+        }
         s#^\.?/##;
         warning "$dir/package.json#files: $_ does not exists\n"
-          unless $_ and -e "$dir/$_" or /[\{\*]/;
-        $_ ? "$dir/$_" : ();
+          unless $nowarn
+          or ( $_ and -e "$dir/$_" or /[\{\*]/ );
+        $_ ? "$prefix$dir/$_" : ();
     } @lines;
     push @files, "$dir/package.json"
-      unless grep { $_ eq 'package.json' } @files;
+      unless grep { $_ eq "$dir/package.json" } @files;
     return @files;
 }
 
@@ -589,44 +623,98 @@ sub configure {
       unless -e 'node_modules';
 
     # External links
-    foreach my $dir ( $self->main_package, @dirs ) {
+    foreach my $dir (
+        (
+            $self->main_package
+            ? (
+                $self->main_package ne '.'
+                ? ( $self->main_package, '.' )
+                : ('.')
+            )
+            : ('.')
+        ),
+        @dirs
+      )
+    {
         if ( -e "debian/nodejs/$dir/extlinks" ) {
             map {
                 my ( $in, $src );
-                run [ 'nodepath', $_ ], \$in, \$src;
-                if ( $? or not $src ) {
-                    print STDERR "### $_ NOT FOUND\n";
+                $_ =~ s/\s+(\S*).*?$//;
+                if (
+                        $1
+                    and $1 eq 'test'
+                    and (
+                        (
+                                $ENV{DEB_BUILD_OPTIONS}
+                            and $ENV{DEB_BUILD_OPTIONS} =~ /\bnocheck\b/
+                        )
+                        or (    $ENV{DEB_BUILD_PROFILES}
+                            and $ENV{DEB_BUILD_PROFILES} =~ /\bnocheck\b/ )
+                    )
+                  )
+                {
+                    print STDERR "Skipping $_ link\n";
                 }
                 else {
-                    chomp $src;
-                    my $tmp = $src;
-                    $tmp =~ s#^.*?nodejs/##;
-                    $tmp =~ s#/.*?$## or $tmp = '';
-                    $tmp = "$tmp/" if $tmp;
-                    $self->dolink( $src, "$dir/node_modules/$tmp" );
+                    run [ 'nodepath', $_ ], \$in, \$src;
+                    if ( $? or not $src ) {
+                        print STDERR
+                          "### $_ is required by debian/nodejs/$dir/extlinks"
+                          . " but not available\n";
+                        exit 1;
+                    }
+                    else {
+                        chomp $src;
+                        my $tmp = $src;
+                        $tmp =~ s#^.*?nodejs/##;
+                        $tmp =~ s#/.*?$## or $tmp = '';
+                        $tmp = "$tmp/" if $tmp;
+                        $self->dolink( $src, "$dir/node_modules/$tmp" );
+                    }
                 }
             } $self->open_file("debian/nodejs/$dir/extlinks");
         }
         if ( -e "debian/nodejs/$dir/extcopies" ) {
             map {
                 my ( $in, $src );
-                run [ 'nodepath', $_ ], \$in, \$src;
-                if ( $? or not $src ) {
-                    print STDERR "### $_ NOT FOUND\n";
+                $_ =~ s/\s+(\S*).*?$//;
+                if (
+                        $1
+                    and $1 eq 'test'
+                    and (
+                        (
+                                $ENV{DEB_BUILD_OPTIONS}
+                            and $ENV{DEB_BUILD_OPTIONS} =~ /\bnocheck\b/
+                        )
+                        or (    $ENV{DEB_BUILD_PROFILES}
+                            and $ENV{DEB_BUILD_PROFILES} =~ /\bnocheck\b/ )
+                    )
+                  )
+                {
+                    print STDERR "Skipping $_ copy\n";
                 }
                 else {
-                    chomp $src;
-                    my $dst = $src;
-                    $dst =~ s#.*?nodejs/##;
-                    if ( $dst =~ s#/[^/]*$## ) {
-                        $self->doit_in_builddir( 'mkdir', '-p',
-                            "$dir/node_modules/$dst" );
+                    run [ 'nodepath', $_ ], \$in, \$src;
+                    if ( $? or not $src ) {
+                        print STDERR
+                          "### $_ is required by debian/nodejs/$dir/extcopies"
+                          . " but not available\n";
+                        exit 1;
                     }
                     else {
-                        $dst = '';
+                        chomp $src;
+                        my $dst = $src;
+                        $dst =~ s#.*?nodejs/##;
+                        if ( $dst =~ s#/[^/]*$## ) {
+                            $self->doit_in_builddir( 'mkdir', '-p',
+                                "$dir/node_modules/$dst" );
+                        }
+                        else {
+                            $dst = '';
+                        }
+                        $self->doit_in_builddir( 'cp', '-rL', $src,
+                            "$dir/node_modules/$dst" );
                     }
-                    $self->doit_in_builddir( 'cp', '-r', $src,
-                        "$dir/node_modules/$dst" );
                 }
             } $self->open_file("debian/nodejs/$dir/extcopies");
         }
@@ -858,14 +946,34 @@ EOT
             my $tmp = 'node-' . $self->pjson($dir)->{name};
             $tmp =~ s/[\@_\/]/-/g;
             $tmp =~ s/--*/-/g;
-            if ( my $v = $self->pjson($dir)->{version} ) {
+            my $v;
+            if ( defined( $v = $self->pjson($dir)->{version} ) ) {
                 $tmp .= " (= $v)";
             }
             push @provides, $tmp;
         }
     }
-    addsubstvar( $package, 'nodejs:Provides', join( ', ', @provides ) )
-      if @provides;
+    my $pname = "node-$node_module";
+    $pname =~ s/[\@_\/]/-/g;
+    $pname =~ s/--*/-/g;
+    if ( $package ne $pname ) {
+        my $tmp = $pname;
+        my $v;
+        if ( -e $self->main_package . "/package.json"
+            and defined( $v = $self->pjson( $self->main_package )->{version} ) )
+        {
+            $tmp .= " (= $v)";
+        }
+        else {
+            print STDERR "# No version found in main package!";
+        }
+        unshift @provides, $tmp;
+    }
+    if (@provides) {
+        print "Populate \${nodejs:Provides}:\n";
+        print " + $_\n" foreach (@provides);
+        addsubstvar( $package, 'nodejs:Provides', join( ', ', @provides ) );
+    }
 
     {
         my ( $in, $src );
@@ -936,11 +1044,11 @@ sub clean {
             $self->do_unlink("$dest/node_modules/$components->{$src}");
         } $self->open_file('debian/nodejs/component_links');
     }
-    foreach my $component (@dirs) {
+    foreach my $component ( $self->main_package, @dirs ) {
         my $package = $components->{$component};
 
         # Error not catched here
-        $self->do_unlink("node_modules/$package");
+        $self->do_unlink("node_modules/$package") if $package;
     }
     if ( -d 'debian/build_modules' ) {
         opendir my $buildDir, 'debian/build_modules';
